@@ -1,3 +1,4 @@
+import ast
 import logging
 import os
 import typing as t
@@ -17,12 +18,9 @@ from blendsql.models.utils import user
 from blendsql.models.constrained.utils import LMString, maybe_load_lm
 from blendsql.ingredients.ingredient import MapIngredient
 from blendsql.common.exceptions import IngredientException
-from blendsql.ingredients.utils import (
-    initialize_retriever,
-    partialclass,
-)
+from blendsql.ingredients.utils import initialize_retriever, partialclass, gen_list
 from blendsql.configure import MAX_OPTIONS_IN_PROMPT_KEY, DEFAULT_MAX_OPTIONS_IN_PROMPT
-from blendsql.types import DataType, prepare_datatype
+from blendsql.types import DataType, prepare_datatype, QuantifierType
 from .examples import (
     MapExample,
     AnnotatedMapExample,
@@ -180,6 +178,7 @@ class LLMMap(MapIngredient):
         ] = None,
         value_limit: t.Optional[int] = None,
         example_outputs: t.Optional[str] = None,
+        quantifier: QuantifierType = None,
         return_type: t.Optional[t.Union[DataType, str]] = None,
         regex: t.Optional[str] = None,
         context: t.Optional[pd.DataFrame] = None,
@@ -207,38 +206,57 @@ class LLMMap(MapIngredient):
         if few_shot_retriever is None:
             few_shot_retriever = lambda *_: DEFAULT_MAP_FEW_SHOT
 
+        # Resolve context argument
+        # If we explicitly passed `context`, this should take precedence over the vector store.
         context_in_use: t.List[str] = [None] * len(values)
         context_in_use_type: ContextType = None
         # If we explicitly passed `context`, this should take precedence over the vector store.
         if searcher is not None and context is None:
             if unpacked_questions:  # Implies we have different context for each value
                 context_in_use = searcher(unpacked_questions)
-                logger.debug(
-                    Fore.LIGHTBLACK_EX
-                    + f"Retrieved contexts '{[str(d[:2]) + '...' for d in context_in_use[:3]]}...'"
-                    + Fore.RESET
-                )
                 context_in_use_type = ContextType.LOCAL
             else:
                 context_in_use = " | ".join(searcher(question)[0])
-                logger.debug(
-                    Fore.LIGHTBLACK_EX
-                    + f"Retrieved context '{context_in_use[:50]}...'"
-                    + Fore.RESET
-                )
                 context_in_use_type = ContextType.GLOBAL
         elif context is not None:  # If we've passed a table context
-            if isinstance(context, pd.DataFrame):
-                context_in_use = context_formatter(context)
-            context_in_use_type = ContextType.GLOBAL
+            if all([len(c) == 1 for c in context]):
+                context_in_use = " | ".join([context_formatter(c) for c in context])
+                context_in_use_type = ContextType.GLOBAL
+            else:
+                assert all(
+                    len(x) == len(values) for x in context
+                ), "Length of scalars passed in LLMMap `context` doesn't match number of values!"
+                context_in_use = [
+                    list(row)
+                    for row in zip(*[df.iloc[:, 0].tolist() for df in context])
+                ]  # Kind of ugly
+                context_in_use_type = ContextType.LOCAL
+
+        # Log what we found
+        if context_in_use_type == ContextType.GLOBAL:
+            logger.debug(
+                Fore.LIGHTBLACK_EX
+                + f"Retrieved global context '{context_in_use[:50]}...'"
+                + Fore.RESET
+            )
+        elif context_in_use_type == ContextType.LOCAL:
+            logger.debug(
+                Fore.LIGHTBLACK_EX
+                + f"Retrieved local contexts '{[str(d[:2]) + '...' for d in context_in_use[:3]]}...'"
+                + Fore.RESET
+            )
+        elif context_in_use_type is not None:
+            raise ValueError(
+                f"Invalid `context_in_use_type`: {type(context_in_use_type)}"
+            )
 
         # Unpack default kwargs
         table_name, column_name = self.unpack_default_kwargs(**kwargs)
         if value_limit is not None:
             values = values[:value_limit]
-        values = [value if not pd.isna(value) else "-" for value in values]
+        values = [str(value) if not pd.isna(value) else "-" for value in values]
         resolved_return_type: DataType = prepare_datatype(
-            return_type=return_type, options=options, quantifier=None
+            return_type=return_type, options=options, quantifier=quantifier
         )
         current_example = MapExample(
             question=question,
@@ -267,7 +285,9 @@ class LLMMap(MapIngredient):
                 )
             ]
 
+        is_list_output = resolved_return_type.quantifier is not None
         regex = regex or resolved_return_type.regex
+        quantifier = resolved_return_type.quantifier
 
         if options is not None and list_options_in_prompt:
             if len(options) > int(
@@ -288,31 +308,38 @@ class LLMMap(MapIngredient):
                 )
 
             lm = LMString()  # type: ignore
-
-            if options is not None and self.enable_constrained_decoding:
-                gen_f = lambda _: guidance.select(options=options)  # type: ignore
-            elif (
-                resolved_return_type.name == "substring"
-                and self.enable_constrained_decoding
-            ):
-                # Special case for substring datatypes
-                gen_f = lambda s: guidance.substring(target_string=s)
-            else:
-                if not self.enable_constrained_decoding:
-                    logger.debug(
-                        Fore.YELLOW
-                        + "Not applying constraints, since `enable_constrained_decoding==False`"
-                        + Fore.RESET
-                    )
-                gen_f = lambda _: guidance.gen(
-                    max_tokens=kwargs.get("max_tokens", 200),
-                    # guidance=0.2.1 doesn't allow both `stop` and `regex` to be passed
-                    stop=None
-                    if regex is not None
-                    else [")", f"\n{INDENT()}"]
-                    + (['"'] if resolved_return_type.name == "str" else []),
+            if is_list_output and self.enable_constrained_decoding:
+                gen_f = lambda _: gen_list(
+                    force_quotes=bool("str" in resolved_return_type.name),
+                    quantifier=quantifier,
+                    options=options,
                     regex=regex,
-                )  # type: ignore
+                )
+            else:
+                if options is not None and self.enable_constrained_decoding:
+                    gen_f = lambda _: guidance.select(options=options)  # type: ignore
+                elif (
+                    resolved_return_type.name == "substring"
+                    and self.enable_constrained_decoding
+                ):
+                    # Special case for substring datatypes
+                    gen_f = lambda s: guidance.substring(target_string=s)
+                else:
+                    if not self.enable_constrained_decoding:
+                        logger.debug(
+                            Fore.YELLOW
+                            + "Not applying constraints, since `enable_constrained_decoding==False`"
+                            + Fore.RESET
+                        )
+                    gen_f = lambda _: guidance.gen(
+                        max_tokens=kwargs.get("max_tokens", 200),
+                        # guidance=0.2.1 doesn't allow both `stop` and `regex` to be passed
+                        stop=None
+                        if regex is not None
+                        else [")", f"\n{INDENT()}"]
+                        + (['"'] if resolved_return_type.name == "str" else []),
+                        regex=regex if self.enable_constrained_decoding else None,
+                    )  # type: ignore
 
             def make_prediction(
                 value: str,
@@ -341,7 +368,7 @@ class LLMMap(MapIngredient):
                     gen_str = (
                         f"""{INDENT(2)}f({value_quote}{indented_value}{value_quote})"""
                     )
-                gen_str += f""" == {'"' if str_output else ''}{guidance.capture(gen_f(value), name=value)}{'"' if str_output else ''}"""
+                gen_str += f""" == {'"' if str_output else ''}{guidance.capture(gen_f(value), name=f"{value}_{context}" if context is not None else value)}{'"' if str_output else ''}"""
                 return gen_str
 
             example_str = ""
@@ -351,9 +378,16 @@ class LLMMap(MapIngredient):
 
             loaded_lm = False
             batch_inference_strings = []
-            batch_inference_values = []
-            value_to_cache_key = {}
+            batch_inference_identifiers = []
+            all_identifiers = []
+            identifier_to_cache_key = {}
             for c, v in zip(context_in_use, values):
+                if c is None:
+                    curr_identifier = v
+                else:
+                    curr_identifier = f"{v}_{c}"
+                all_identifiers.append(curr_identifier)
+
                 if context_in_use_type == ContextType.LOCAL:
                     current_example.context = c
 
@@ -370,16 +404,19 @@ class LLMMap(MapIngredient):
                         example_str,
                         current_example_str,
                         question,
+                        regex,
                         options,
+                        quantifier,
+                        kwargs.get("max_tokens", 200),
                         c,
                         v,
                         funcs=[make_prediction, gen_f],
                     )
                     if cached_response is not None:
-                        lm = lm.set(v, cached_response)
+                        lm = lm.set(curr_identifier, cached_response)
                         in_cache = True
                     else:
-                        value_to_cache_key[v] = cache_key
+                        identifier_to_cache_key[curr_identifier] = cache_key
 
                 if not in_cache and not loaded_lm:
                     lm: guidance.models.Model = maybe_load_lm(model, lm)
@@ -400,7 +437,7 @@ class LLMMap(MapIngredient):
                             gen_f=gen_f,
                         )
                     )
-                    batch_inference_values.append(v)
+                    batch_inference_identifiers.append(curr_identifier)
 
             with guidance.assistant():
                 iter = range(0, len(batch_inference_strings), batch_size)
@@ -421,27 +458,47 @@ class LLMMap(MapIngredient):
                     )
                     add_to_global_history(str(batch_lm))
                     if model.caching:
-                        for value in batch_inference_values[i : i + batch_size]:
-                            cache_key = value_to_cache_key[value]
-                            model.cache[cache_key] = lm.get(value)  # type: ignore
+                        for identifier in batch_inference_identifiers[
+                            i : i + batch_size
+                        ]:
+                            cache_key = identifier_to_cache_key[identifier]
+                            model.cache[cache_key] = lm.get(identifier)  # type: ignore
 
-            lm_mapping: t.List[str] = [lm[value] for value in values]  # type: ignore
+            lm_mapping: t.List[str] = [lm[identifier] for identifier in all_identifiers]  # type: ignore
             model.completion_tokens += sum(
                 [len(model.tokenizer.encode(v)) for v in lm_mapping]
             )
             model.prompt_tokens += lm._get_usage().input_tokens
             # For each value, call the DataType's `coerce_fn()`
-            mapped_values = [resolved_return_type.coerce_fn(s) for s in lm_mapping]
+            if is_list_output:
+                mapped_values = [
+                    [
+                        current_example.return_type.coerce_fn(c)
+                        for c in ast.literal_eval(s)
+                    ]
+                    for s in lm_mapping
+                ]
+            else:
+                mapped_values = [resolved_return_type.coerce_fn(s) for s in lm_mapping]
+
         else:
-            sorted_values = sorted(values)
+            sorted_indices = sorted(range(len(values)), key=lambda i: values[i])
+            sorted_indices_to_original = {
+                k: idx for idx, k in enumerate(sorted_indices)
+            }
+            sorted_values = [values[i] for i in sorted_indices]
+            if context_in_use_type is not None:
+                context_in_use = [context_in_use[i] for i in sorted_indices]
+
             messages_list: t.List[t.List[dict]] = []
             batch_sizes: t.List[int] = []
             if current_example.context_type == ContextType.LOCAL:
-                logger.debug(
-                    Fore.YELLOW
-                    + f"Overriding batch_size={batch_size} to 0, since UnconstrainedModels with LLMMap don't support local context for now"
-                    + Fore.RESET
-                )
+                if batch_size != 1:
+                    logger.debug(
+                        Fore.YELLOW
+                        + f"Overriding batch_size={batch_size} to 1, since UnconstrainedModels with LLMMap don't support local context for now"
+                        + Fore.RESET
+                    )
                 batch_size = 1
                 current_example.context_type = ContextType.GLOBAL
                 current_example.context = None
@@ -457,7 +514,8 @@ class LLMMap(MapIngredient):
                     user_msg_str += example.to_string()
                 # Add the current question + context for inference
                 if current_batch_example.context_type == ContextType.GLOBAL:
-                    current_batch_example.context = "\n".join(curr_batch_contexts[0])
+                    str_context = "\n".join(curr_batch_contexts[0])
+                    current_batch_example.context = str_context
                 user_msg_str += current_batch_example.to_string(
                     values=curr_batch_values,
                     list_options=list_options_in_prompt,
@@ -473,17 +531,38 @@ class LLMMap(MapIngredient):
             total_missing_values = 0
             for idx, r in enumerate(responses):
                 expected_len = batch_sizes[idx]
-                predictions: t.List[Union[str, None]] = r.split(DEFAULT_ANS_SEP)  # type: ignore
+                predictions: t.List[t.Union[str, None]] = r.split(DEFAULT_ANS_SEP)  # type: ignore
+                # Add null values, if we under-predicted
                 while len(predictions) < expected_len:
                     total_missing_values += 1
                     predictions.append(None)
-                # Try to map to booleans, `None`, and numeric datatypes
-                mapped_values.extend(
-                    [current_example.return_type.coerce_fn(s) for s in predictions]
-                )
-
-            mapping = {k: v for k, v in zip(sorted_values, mapped_values)}
-            mapped_values = [mapping[value] for value in values]
+                # Cutoff, if we over-predicted
+                predictions = predictions[:expected_len]
+                if is_list_output:
+                    curr_converted_preds = []
+                    for pred in predictions:
+                        try:
+                            list_converted = ast.literal_eval(pred)
+                        except (ValueError, SyntaxError):
+                            logger.debug(
+                                Fore.RED
+                                + f"Error casting prediction '{pred}' to a list"
+                            )
+                            curr_converted_preds.append([])
+                            continue
+                        for item in list_converted:
+                            curr_converted_preds.append(
+                                current_example.return_type.coerce_fn(item)
+                            )
+                    mapped_values.append(curr_converted_preds)
+                else:
+                    mapped_values.extend(
+                        [current_example.return_type.coerce_fn(s) for s in predictions]
+                    )
+            mapped_values = [
+                mapped_values[sorted_indices_to_original[i]]
+                for i in range(len(mapped_values))
+            ]
 
             if total_missing_values > 0:
                 logger.debug(
