@@ -1,6 +1,5 @@
 import os
 import copy
-import re
 from ast import literal_eval
 from pathlib import Path
 import typing as t
@@ -18,14 +17,11 @@ from blendsql.models.constrained.utils import maybe_load_lm, LMString
 from blendsql.ingredients.ingredient import QAIngredient
 from blendsql.db.utils import single_quote_escape
 from blendsql.common.exceptions import IngredientException
-from blendsql.ingredients.utils import (
-    initialize_retriever,
-    partialclass,
-)
+from blendsql.ingredients.utils import initialize_retriever, partialclass, gen_list
 from blendsql.configure import MAX_OPTIONS_IN_PROMPT_KEY, DEFAULT_MAX_OPTIONS_IN_PROMPT
 from blendsql.types import DataType, QuantifierType, prepare_datatype
-from .examples import QAExample, AnnotatedQAExample
 from blendsql.search.searcher import Searcher
+from .examples import QAExample, AnnotatedQAExample
 
 MAIN_INSTRUCTION = "Answer the question given the context, if provided.\n"
 LONG_ANSWER_INSTRUCTION = "Make the answer as concrete as possible, providing more context and reasoning using the entire context.\n"
@@ -175,7 +171,7 @@ class LLMQA(QAIngredient):
         quantifier: QuantifierType = None,
         return_type: t.Optional[t.Union[DataType, str]] = None,
         regex: t.Optional[str] = None,
-        context: t.List[pd.DataFrame] = None,
+        context: t.Optional[t.List[pd.DataFrame]] = None,
         long_answer: bool = False,
         use_option_aliases: bool = False,
         **kwargs,
@@ -260,57 +256,6 @@ class LLMQA(QAIngredient):
         if isinstance(model, ConstrainedModel):
             import guidance
 
-            def get_quantifier_wrapper(
-                quantifier: QuantifierType,
-            ) -> t.Callable[[guidance.models.Model], guidance.models.Model]:
-                quantifier_wrapper = lambda x: x
-                if quantifier is not None:
-                    if quantifier == "*":
-                        quantifier_wrapper = guidance.zero_or_more
-                    elif quantifier == "+":
-                        quantifier_wrapper = guidance.one_or_more
-                    elif re.match(r"{\d+}", quantifier):
-                        repeats = [
-                            int(i)
-                            for i in quantifier.replace("}", "")
-                            .replace("{", "")
-                            .split(",")
-                        ]
-                        if len(repeats) == 1:
-                            repeats = repeats * 2
-                        min_length, max_length = repeats
-                        quantifier_wrapper = lambda f: guidance.sequence(
-                            f, min_length=min_length, max_length=max_length
-                        )  # type: ignore
-                return quantifier_wrapper  # type: ignore
-
-            @guidance(stateless=True, dedent=False)
-            def gen_list(
-                lm,
-                force_quotes: bool,
-                quantifier=None,
-                options: t.Optional[t.List[str]] = None,
-                regex: t.Optional[str] = None,
-            ):
-                if options:
-                    single_item = guidance.select(
-                        options, list_append=True, name="response"
-                    )
-                else:
-                    single_item = guidance.gen(
-                        max_tokens=100,
-                        # If not regex is passed, default to all characters except these specific to list-syntax
-                        regex=regex or "[^],']+",
-                        list_append=True,
-                        name="response",
-                    )  # type: ignore
-                quote = "'"
-                if not force_quotes:
-                    quote = guidance.optional(quote)  # type: ignore
-                single_item = quote + single_item + quote
-                single_item += guidance.optional(", ")  # type: ignore
-                return lm + "[" + get_quantifier_wrapper(quantifier)(single_item) + "]"
-
             lm = LMString()  # type: ignore
 
             instruction_str = MAIN_INSTRUCTION
@@ -324,19 +269,15 @@ class LLMQA(QAIngredient):
             )
 
             if is_list_output and self.enable_constrained_decoding:
-                gen_kwargs = {
-                    "force_quotes": bool("str" in resolved_return_type.name),
-                    "regex": regex,
-                    "options": options_with_aliases,
-                    "quantifier": quantifier,
-                }
-                gen_f = gen_list
+                gen_f = lambda _: gen_list(
+                    force_quotes=bool("str" in resolved_return_type.name),
+                    regex=regex,
+                    options=options_with_aliases,
+                    quantifier=quantifier,
+                )
             else:
                 if options and self.enable_constrained_decoding:
-                    # Too many options here raises:
-                    # ValueError: Parser Error: Current row has 10850 items; max is 2000; consider making your grammar left-recursive if it's right-recursive
-                    gen_kwargs = {"options": options, "name": "response"}
-                    gen_f = guidance.select
+                    gen_f = lambda _: guidance.select(options=options, name="response")
                 else:
                     if not self.enable_constrained_decoding:
                         logger.debug(
@@ -344,15 +285,11 @@ class LLMQA(QAIngredient):
                             + "Not applying constraints, since `enable_constrained_decoding==False`"
                             + Fore.RESET
                         )
-                    gen_kwargs = {
-                        "max_tokens": kwargs.get("max_tokens", 200),
-                        "regex": regex if self.enable_constrained_decoding else None,
-                        "name": "response",
-                        # guidance=0.2.1 doesn't allow both `stop` and `regex` to be passed
-                        # "stop": ["\n"] if regex is None else None,
-                    }
-                    gen_f = guidance.gen
-
+                    gen_f = lambda _: guidance.gen(
+                        max_tokens=kwargs.get("max_tokens", 200),
+                        regex=regex if self.enable_constrained_decoding else None,
+                        name="response",
+                    )
             # First check - do we need to load the model?
             in_cache = False
             if model.caching:
@@ -366,7 +303,9 @@ class LLMQA(QAIngredient):
                         ]
                     ),
                     regex,
-                    gen_kwargs,
+                    options,
+                    quantifier,
+                    kwargs.get("max_tokens", 200),
                     funcs=[gen_f],
                 )
                 if response is not None:
@@ -387,7 +326,7 @@ class LLMQA(QAIngredient):
                     lm += curr_example_str
 
                 with guidance.assistant():
-                    lm += gen_f(**gen_kwargs)
+                    lm += gen_f(question)
                 add_to_global_history(str(lm))
 
                 response: str = lm["response"]
