@@ -161,20 +161,6 @@ class Ingredient:
             )
         return list(unpacked_options) if len(unpacked_options) > 0 else None
 
-    def unpack_question(
-        self,
-        question: str,
-        values: t.List[str],
-    ) -> t.List[str]:
-        """Unpack any f-string ValueArray references in question.
-
-        Example:
-            question='Where is {} located?', values=["Sydney", "San Jose"]
-        """
-        if "{}" in question:
-            return [question.format(value) for value in values]
-        return [question for _ in range(len(values))]
-
 
 @attrs
 class AliasIngredient(Ingredient):
@@ -273,12 +259,22 @@ class MapIngredient(Ingredient):
         self,
         question: t.Optional[str] = None,
         values: t.Optional[t.Union[ColumnRef]] = None,
+        *context: t.Union[str, pd.DataFrame],
         options: t.Optional[t.Union[ColumnRef, list]] = None,
-        *args,
         **kwargs,
     ) -> tuple:
         """Returns tuple with format (arg, tablename, colname, new_table)"""
         # Unpack kwargs
+        # Extract single `context` from kwargs if provided
+        if "context" in kwargs:
+            context_kwarg = kwargs.pop("context")
+            # Combine positional and keyword context
+            if isinstance(context_kwarg, (list, tuple)):
+                context = context + tuple(context_kwarg)
+            else:
+                context = context + (context_kwarg,)
+
+        context_was_passed = len(context) > 0
         aliases_to_tablenames: t.Dict[str, str] = kwargs["aliases_to_tablenames"]
         get_temp_subquery_table: t.Callable = kwargs["get_temp_subquery_table"]
         get_temp_session_table: t.Callable = kwargs["get_temp_session_table"]
@@ -314,6 +310,26 @@ class MapIngredient(Ingredient):
         ):
             new_arg_column = "_" + new_arg_column
 
+        if context_was_passed:
+            all_context_colnames = []
+            for _context in context:
+                if isinstance(_context, ColumnRef):
+                    _, context_colname = utils.get_tablename_colname(_context)
+                    all_context_colnames.append(context_colname)
+                else:
+                    raise ValueError(
+                        f"Not sure what to do with context arg passed to MapIngredient: {_context}"
+                    )
+            select_distinct_fn = lambda q: self.db.execute_to_df(q)
+            select_distinct_arg = (
+                f'"{colname}"'
+                + ", "
+                + ", ".join([f'"{c}"' for c in all_context_colnames])
+            )
+        else:
+            select_distinct_fn = lambda q: self.db.execute_to_list(q)
+            select_distinct_arg = f'"{colname}"'
+
         # Get a list of values to map
         # First, check if we've already dumped some `MapIngredient` output to the main session table
         if temp_session_table_exists:
@@ -323,20 +339,31 @@ class MapIngredient(Ingredient):
             # We don't need to run this function on everything,
             #   if a previous subquery already got to certain values
             if new_arg_column in temp_session_table.columns:
-                unpacked_values = self.db.execute_to_list(
-                    f'SELECT DISTINCT "{colname}" FROM "{temp_session_tablename}" WHERE "{new_arg_column}" IS NULL',
+                distinct_values = select_distinct_fn(
+                    f'SELECT DISTINCT {select_distinct_arg} FROM "{temp_session_tablename}" WHERE "{new_arg_column}" IS NULL',
                 )
             # Base case: this is the first time we've used this particular ingredient
             # BUT, temp_session_tablename still exists
             else:
-                unpacked_values = self.db.execute_to_list(
-                    f'SELECT DISTINCT "{colname}" FROM "{temp_session_tablename}"',
+                distinct_values = select_distinct_fn(
+                    f'SELECT DISTINCT {select_distinct_arg} FROM "{temp_session_tablename}"',
                 )
         else:
-            unpacked_values = self.db.execute_to_list(
-                f'SELECT DISTINCT "{colname}" FROM "{value_source_tablename}"',
+            distinct_values = select_distinct_fn(
+                f'SELECT DISTINCT {select_distinct_arg} FROM "{value_source_tablename}"',
             )
 
+        context_subtables = []
+        if context_was_passed:
+            unpacked_values: list = distinct_values[colname].tolist()
+            context_subtables = [
+                pd.DataFrame(distinct_values[c])
+                for c in distinct_values.columns
+                if c != colname
+            ]
+        else:
+            unpacked_values: list = distinct_values
+
         # No need to run ingredient if we have no values to map onto
         if len(unpacked_values) == 0:
             original_table[new_arg_column] = None
@@ -351,10 +378,23 @@ class MapIngredient(Ingredient):
 
         unpacked_questions = None
         if question is not None:
-            if "{}" in question:
+            num_braces = question.count("{}")
+            if num_braces > 0:
+                # # Pop off 'context' into question, if there are > 1 '{}'
+                # lists_to_zip = []
+                # for i in range(num_braces - 1):
+                #     lists_to_zip.append(context_subtables[i].values.flatten().tolist())
+                # extra_values_to_insert = list(zip(*lists_to_zip))
+
                 unpacked_questions = [
                     question.format(value) for value in unpacked_values
                 ]
+
+                # unpacked_questions = [
+                #     question.format(value, *extra_values) for value, extra_values in
+                #     zip(unpacked_values, extra_values_to_insert)
+                # ]
+
                 logger.debug(
                     Fore.LIGHTBLACK_EX
                     + f"Unpacked question to '{unpacked_questions[:10]}'"
@@ -365,10 +405,10 @@ class MapIngredient(Ingredient):
             question=question,
             unpacked_questions=unpacked_questions,
             values=unpacked_values,
+            context=context_subtables if len(context_subtables) > 0 else None,
             options=options,
             tablename=tablename,
             colname=colname,
-            *args,
             **self.__dict__ | kwargs,
         )
         self.num_values_passed += len(mapped_values)
@@ -376,14 +416,23 @@ class MapIngredient(Ingredient):
         for value, mapped_value in zip(unpacked_values, mapped_values):
             df_as_dict[colname].append(value)
             df_as_dict[new_arg_column].append(mapped_value)
-        subtable = pd.DataFrame(df_as_dict)
+        mapped_subtable = pd.DataFrame(df_as_dict)
         if all(
             isinstance(x, (int, type(None))) and not isinstance(x, bool)
             for x in mapped_values
         ):
-            subtable[new_arg_column] = subtable[new_arg_column].astype("Int64")
+            mapped_subtable[new_arg_column] = mapped_subtable[new_arg_column].astype(
+                "Int64"
+            )
         # Add new_table to original table
-        new_table = original_table.merge(subtable, how="left", on=colname)
+        if context_was_passed:
+            _mapped_subtable = distinct_values
+            _mapped_subtable[new_arg_column] = mapped_subtable[new_arg_column]
+            new_table = original_table.merge(
+                _mapped_subtable, how="left", on=[colname] + all_context_colnames
+            )
+        else:
+            new_table = original_table.merge(mapped_subtable, how="left", on=colname)
         if new_table.shape[0] != original_table.shape[0]:
             raise IngredientException(
                 f"subtable from MapIngredient.run() needs same length as # rows from original\nOriginal has {original_table.shape[0]}, new_table has {new_table.shape[0]}"
@@ -471,11 +520,15 @@ class JoinIngredient(Ingredient):
                 )
             )
             modified_lr_identifiers.append((tablename, colname))
-        sorted_values = sorted(values, key=len)
-        # check swapping only once, at the beginning
-        if sorted_values != values:
-            swapped = True
+
+        sorted_values = values
+        swapped = False
         if join_criteria is None:
+            # Only do order optimization if we haven't passed a custom `join_criteria`
+            sorted_values = sorted(values, key=len)
+            # check swapping only once, at the beginning
+            if sorted_values != values:
+                swapped = True
             # First, check which values we actually need to call Model on
             # We don't want to join when there's already an intuitive alignment
             # First, make sure outer loop is shorter of the two lists
@@ -534,6 +587,7 @@ class JoinIngredient(Ingredient):
 
         # Now, we have our final values to process.
         left_values, right_values = sorted_values
+        # right_values, left_values = sorted_values
 
         (left_tablename, left_colname), (
             right_tablename,
